# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/protobuf/clients/clients.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from yamcs.api import annotations_pb2 as yamcs_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/protobuf/clients/clients.proto',
  package='yamcs.protobuf.clients',
  syntax='proto2',
  serialized_options=b'\n\022org.yamcs.protobufB\014ClientsProtoP\001',
  serialized_pb=b'\n$yamcs/protobuf/clients/clients.proto\x12\x16yamcs.protobuf.clients\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1byamcs/api/annotations.proto\"J\n\x13ListClientsResponse\x12\x33\n\x07\x63lients\x18\x01 \x03(\x0b\x32\".yamcs.protobuf.clients.ClientInfo\"\x1e\n\x10GetClientRequest\x12\n\n\x02id\x18\x01 \x01(\x05\"D\n\x11\x45\x64itClientRequest\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x10\n\x08instance\x18\x02 \x01(\t\x12\x11\n\tprocessor\x18\x03 \x01(\t\"\xa7\x02\n\nClientInfo\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x17\n\x0f\x61pplicationName\x18\x04 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\n \x01(\t\x12\x10\n\x08instance\x18\x01 \x01(\t\x12\x15\n\rprocessorName\x18\x05 \x01(\t\x12=\n\x05state\x18\x06 \x01(\x0e\x32..yamcs.protobuf.clients.ClientInfo.ClientState\x12-\n\tloginTime\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\".\n\x0b\x43lientState\x12\r\n\tCONNECTED\x10\x00\x12\x10\n\x0c\x44ISCONNECTED\x10\x01J\x04\x08\x08\x10\tJ\x04\x08\t\x10\n2\xd7\x02\n\nClientsApi\x12\x66\n\x0bListClients\x12\x16.google.protobuf.Empty\x1a+.yamcs.protobuf.clients.ListClientsResponse\"\x12\x8a\x92\x03\x0e\n\x0c/api/clients\x12r\n\tGetClient\x12(.yamcs.protobuf.clients.GetClientRequest\x1a\".yamcs.protobuf.clients.ClientInfo\"\x17\x8a\x92\x03\x13\n\x11/api/clients/{id}\x12m\n\x0cUpdateClient\x12).yamcs.protobuf.clients.EditClientRequest\x1a\x16.google.protobuf.Empty\"\x1a\x8a\x92\x03\x16*\x11/api/clients/{id}:\x01*B$\n\x12org.yamcs.protobufB\x0c\x43lientsProtoP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,yamcs_dot_api_dot_annotations__pb2.DESCRIPTOR,])



_CLIENTINFO_CLIENTSTATE = _descriptor.EnumDescriptor(
  name='ClientState',
  full_name='yamcs.protobuf.clients.ClientInfo.ClientState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CONNECTED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISCONNECTED', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=571,
  serialized_end=617,
)
_sym_db.RegisterEnumDescriptor(_CLIENTINFO_CLIENTSTATE)


_LISTCLIENTSRESPONSE = _descriptor.Descriptor(
  name='ListClientsResponse',
  full_name='yamcs.protobuf.clients.ListClientsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clients', full_name='yamcs.protobuf.clients.ListClientsResponse.clients', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=229,
)


_GETCLIENTREQUEST = _descriptor.Descriptor(
  name='GetClientRequest',
  full_name='yamcs.protobuf.clients.GetClientRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.clients.GetClientRequest.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=231,
  serialized_end=261,
)


_EDITCLIENTREQUEST = _descriptor.Descriptor(
  name='EditClientRequest',
  full_name='yamcs.protobuf.clients.EditClientRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.clients.EditClientRequest.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.clients.EditClientRequest.instance', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='processor', full_name='yamcs.protobuf.clients.EditClientRequest.processor', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=263,
  serialized_end=331,
)


_CLIENTINFO = _descriptor.Descriptor(
  name='ClientInfo',
  full_name='yamcs.protobuf.clients.ClientInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.clients.ClientInfo.id', index=0,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='username', full_name='yamcs.protobuf.clients.ClientInfo.username', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='applicationName', full_name='yamcs.protobuf.clients.ClientInfo.applicationName', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='yamcs.protobuf.clients.ClientInfo.address', index=3,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='instance', full_name='yamcs.protobuf.clients.ClientInfo.instance', index=4,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='processorName', full_name='yamcs.protobuf.clients.ClientInfo.processorName', index=5,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='yamcs.protobuf.clients.ClientInfo.state', index=6,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loginTime', full_name='yamcs.protobuf.clients.ClientInfo.loginTime', index=7,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CLIENTINFO_CLIENTSTATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=334,
  serialized_end=629,
)

_LISTCLIENTSRESPONSE.fields_by_name['clients'].message_type = _CLIENTINFO
_CLIENTINFO.fields_by_name['state'].enum_type = _CLIENTINFO_CLIENTSTATE
_CLIENTINFO.fields_by_name['loginTime'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CLIENTINFO_CLIENTSTATE.containing_type = _CLIENTINFO
DESCRIPTOR.message_types_by_name['ListClientsResponse'] = _LISTCLIENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetClientRequest'] = _GETCLIENTREQUEST
DESCRIPTOR.message_types_by_name['EditClientRequest'] = _EDITCLIENTREQUEST
DESCRIPTOR.message_types_by_name['ClientInfo'] = _CLIENTINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListClientsResponse = _reflection.GeneratedProtocolMessageType('ListClientsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTCLIENTSRESPONSE,
  '__module__' : 'yamcs.protobuf.clients.clients_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.clients.ListClientsResponse)
  })
_sym_db.RegisterMessage(ListClientsResponse)

GetClientRequest = _reflection.GeneratedProtocolMessageType('GetClientRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCLIENTREQUEST,
  '__module__' : 'yamcs.protobuf.clients.clients_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.clients.GetClientRequest)
  })
_sym_db.RegisterMessage(GetClientRequest)

EditClientRequest = _reflection.GeneratedProtocolMessageType('EditClientRequest', (_message.Message,), {
  'DESCRIPTOR' : _EDITCLIENTREQUEST,
  '__module__' : 'yamcs.protobuf.clients.clients_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.clients.EditClientRequest)
  })
_sym_db.RegisterMessage(EditClientRequest)

ClientInfo = _reflection.GeneratedProtocolMessageType('ClientInfo', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTINFO,
  '__module__' : 'yamcs.protobuf.clients.clients_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.clients.ClientInfo)
  })
_sym_db.RegisterMessage(ClientInfo)


DESCRIPTOR._options = None

_CLIENTSAPI = _descriptor.ServiceDescriptor(
  name='ClientsApi',
  full_name='yamcs.protobuf.clients.ClientsApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=632,
  serialized_end=975,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListClients',
    full_name='yamcs.protobuf.clients.ClientsApi.ListClients',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LISTCLIENTSRESPONSE,
    serialized_options=b'\212\222\003\016\n\014/api/clients',
  ),
  _descriptor.MethodDescriptor(
    name='GetClient',
    full_name='yamcs.protobuf.clients.ClientsApi.GetClient',
    index=1,
    containing_service=None,
    input_type=_GETCLIENTREQUEST,
    output_type=_CLIENTINFO,
    serialized_options=b'\212\222\003\023\n\021/api/clients/{id}',
  ),
  _descriptor.MethodDescriptor(
    name='UpdateClient',
    full_name='yamcs.protobuf.clients.ClientsApi.UpdateClient',
    index=2,
    containing_service=None,
    input_type=_EDITCLIENTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\212\222\003\026*\021/api/clients/{id}:\001*',
  ),
])
_sym_db.RegisterServiceDescriptor(_CLIENTSAPI)

DESCRIPTOR.services_by_name['ClientsApi'] = _CLIENTSAPI

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/protobuf/web/general_service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from yamcs.api import annotations_pb2 as yamcs_dot_api_dot_annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/protobuf/web/general_service.proto',
  package='yamcs.protobuf.web',
  syntax='proto2',
  serialized_options=b'\n\022org.yamcs.protobufB\023GeneralServiceProtoP\001',
  serialized_pb=b'\n(yamcs/protobuf/web/general_service.proto\x12\x12yamcs.protobuf.web\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1byamcs/api/annotations.proto\"C\n\x12ListRoutesResponse\x12-\n\x06routes\x18\x01 \x03(\x0b\x32\x1d.yamcs.protobuf.web.RouteInfo\"^\n\x1dListClientConnectionsResponse\x12=\n\x0b\x63onnections\x18\x01 \x03(\x0b\x32(.yamcs.protobuf.web.ClientConnectionInfo\"$\n\x16\x43loseConnectionRequest\x12\n\n\x02id\x18\x01 \x01(\t\"\xc7\x01\n\tRouteInfo\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x12\n\nhttpMethod\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x11\n\tinputType\x18\x06 \x01(\t\x12\x12\n\noutputType\x18\x07 \x01(\t\x12\x12\n\ndeprecated\x18\x08 \x01(\x08\x12\x14\n\x0crequestCount\x18\t \x01(\x03\x12\x12\n\nerrorCount\x18\n \x01(\x03\"\x8a\x02\n\x16GetGeneralInfoResponse\x12\x14\n\x0cyamcsVersion\x18\x01 \x01(\t\x12\x10\n\x08revision\x18\x06 \x01(\t\x12\x10\n\x08serverId\x18\x02 \x01(\t\x12\x1c\n\x14\x64\x65\x66\x61ultYamcsInstance\x18\x03 \x01(\t\x12\x46\n\x07plugins\x18\x05 \x03(\x0b\x32\x35.yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo\x1aP\n\nPluginInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0e\n\x06vendor\x18\x04 \x01(\t\"\xfa\x02\n\x14\x43lientConnectionInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04open\x18\x02 \x01(\x08\x12\x0e\n\x06\x61\x63tive\x18\x03 \x01(\x08\x12\x10\n\x08writable\x18\x04 \x01(\x08\x12\x15\n\rremoteAddress\x18\x05 \x01(\t\x12\x11\n\treadBytes\x18\x06 \x01(\x04\x12\x14\n\x0cwrittenBytes\x18\x07 \x01(\x04\x12\x16\n\x0ereadThroughput\x18\x08 \x01(\x04\x12\x17\n\x0fwriteThroughput\x18\t \x01(\x04\x12M\n\x0bhttpRequest\x18\n \x01(\x0b\x32\x38.yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo\x1a\x66\n\x0fHttpRequestInfo\x12\x10\n\x08protocol\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0b\n\x03uri\x18\x03 \x01(\t\x12\x11\n\tkeepAlive\x18\x04 \x01(\x08\x12\x11\n\tuserAgent\x18\x05 \x01(\t2\xbf\x03\n\nGeneralApi\x12`\n\x0eGetGeneralInfo\x12\x16.google.protobuf.Empty\x1a*.yamcs.protobuf.web.GetGeneralInfoResponse\"\n\x8a\x92\x03\x06\n\x04/api\x12_\n\nListRoutes\x12\x16.google.protobuf.Empty\x1a&.yamcs.protobuf.web.ListRoutesResponse\"\x11\x8a\x92\x03\r\n\x0b/api/routes\x12z\n\x15ListClientConnections\x12\x16.google.protobuf.Empty\x1a\x31.yamcs.protobuf.web.ListClientConnectionsResponse\"\x16\x8a\x92\x03\x12\n\x10/api/connections\x12r\n\x0f\x43loseConnection\x12*.yamcs.protobuf.web.CloseConnectionRequest\x1a\x16.google.protobuf.Empty\"\x1b\x8a\x92\x03\x17\"\x15/api/connections/{id}B+\n\x12org.yamcs.protobufB\x13GeneralServiceProtoP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,yamcs_dot_api_dot_annotations__pb2.DESCRIPTOR,])




_LISTROUTESRESPONSE = _descriptor.Descriptor(
  name='ListRoutesResponse',
  full_name='yamcs.protobuf.web.ListRoutesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='routes', full_name='yamcs.protobuf.web.ListRoutesResponse.routes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=122,
  serialized_end=189,
)


_LISTCLIENTCONNECTIONSRESPONSE = _descriptor.Descriptor(
  name='ListClientConnectionsResponse',
  full_name='yamcs.protobuf.web.ListClientConnectionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='connections', full_name='yamcs.protobuf.web.ListClientConnectionsResponse.connections', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=191,
  serialized_end=285,
)


_CLOSECONNECTIONREQUEST = _descriptor.Descriptor(
  name='CloseConnectionRequest',
  full_name='yamcs.protobuf.web.CloseConnectionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.CloseConnectionRequest.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=287,
  serialized_end=323,
)


_ROUTEINFO = _descriptor.Descriptor(
  name='RouteInfo',
  full_name='yamcs.protobuf.web.RouteInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='yamcs.protobuf.web.RouteInfo.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='yamcs.protobuf.web.RouteInfo.method', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='yamcs.protobuf.web.RouteInfo.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='httpMethod', full_name='yamcs.protobuf.web.RouteInfo.httpMethod', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='yamcs.protobuf.web.RouteInfo.url', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inputType', full_name='yamcs.protobuf.web.RouteInfo.inputType', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='outputType', full_name='yamcs.protobuf.web.RouteInfo.outputType', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deprecated', full_name='yamcs.protobuf.web.RouteInfo.deprecated', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requestCount', full_name='yamcs.protobuf.web.RouteInfo.requestCount', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='errorCount', full_name='yamcs.protobuf.web.RouteInfo.errorCount', index=9,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=525,
)


_GETGENERALINFORESPONSE_PLUGININFO = _descriptor.Descriptor(
  name='PluginInfo',
  full_name='yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vendor', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo.vendor', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=714,
  serialized_end=794,
)

_GETGENERALINFORESPONSE = _descriptor.Descriptor(
  name='GetGeneralInfoResponse',
  full_name='yamcs.protobuf.web.GetGeneralInfoResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='yamcsVersion', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.yamcsVersion', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='revision', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.revision', index=1,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='serverId', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.serverId', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='defaultYamcsInstance', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.defaultYamcsInstance', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='plugins', full_name='yamcs.protobuf.web.GetGeneralInfoResponse.plugins', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GETGENERALINFORESPONSE_PLUGININFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=528,
  serialized_end=794,
)


_CLIENTCONNECTIONINFO_HTTPREQUESTINFO = _descriptor.Descriptor(
  name='HttpRequestInfo',
  full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='protocol', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.protocol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.method', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uri', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keepAlive', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.keepAlive', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='userAgent', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.userAgent', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1073,
  serialized_end=1175,
)

_CLIENTCONNECTIONINFO = _descriptor.Descriptor(
  name='ClientConnectionInfo',
  full_name='yamcs.protobuf.web.ClientConnectionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.ClientConnectionInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='open', full_name='yamcs.protobuf.web.ClientConnectionInfo.open', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='active', full_name='yamcs.protobuf.web.ClientConnectionInfo.active', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writable', full_name='yamcs.protobuf.web.ClientConnectionInfo.writable', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='remoteAddress', full_name='yamcs.protobuf.web.ClientConnectionInfo.remoteAddress', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readBytes', full_name='yamcs.protobuf.web.ClientConnectionInfo.readBytes', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writtenBytes', full_name='yamcs.protobuf.web.ClientConnectionInfo.writtenBytes', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readThroughput', full_name='yamcs.protobuf.web.ClientConnectionInfo.readThroughput', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writeThroughput', full_name='yamcs.protobuf.web.ClientConnectionInfo.writeThroughput', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='httpRequest', full_name='yamcs.protobuf.web.ClientConnectionInfo.httpRequest', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CLIENTCONNECTIONINFO_HTTPREQUESTINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=797,
  serialized_end=1175,
)

_LISTROUTESRESPONSE.fields_by_name['routes'].message_type = _ROUTEINFO
_LISTCLIENTCONNECTIONSRESPONSE.fields_by_name['connections'].message_type = _CLIENTCONNECTIONINFO
_GETGENERALINFORESPONSE_PLUGININFO.containing_type = _GETGENERALINFORESPONSE
_GETGENERALINFORESPONSE.fields_by_name['plugins'].message_type = _GETGENERALINFORESPONSE_PLUGININFO
_CLIENTCONNECTIONINFO_HTTPREQUESTINFO.containing_type = _CLIENTCONNECTIONINFO
_CLIENTCONNECTIONINFO.fields_by_name['httpRequest'].message_type = _CLIENTCONNECTIONINFO_HTTPREQUESTINFO
DESCRIPTOR.message_types_by_name['ListRoutesResponse'] = _LISTROUTESRESPONSE
DESCRIPTOR.message_types_by_name['ListClientConnectionsResponse'] = _LISTCLIENTCONNECTIONSRESPONSE
DESCRIPTOR.message_types_by_name['CloseConnectionRequest'] = _CLOSECONNECTIONREQUEST
DESCRIPTOR.message_types_by_name['RouteInfo'] = _ROUTEINFO
DESCRIPTOR.message_types_by_name['GetGeneralInfoResponse'] = _GETGENERALINFORESPONSE
DESCRIPTOR.message_types_by_name['ClientConnectionInfo'] = _CLIENTCONNECTIONINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListRoutesResponse = _reflection.GeneratedProtocolMessageType('ListRoutesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTROUTESRESPONSE,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListRoutesResponse)
  })
_sym_db.RegisterMessage(ListRoutesResponse)

ListClientConnectionsResponse = _reflection.GeneratedProtocolMessageType('ListClientConnectionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTCLIENTCONNECTIONSRESPONSE,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListClientConnectionsResponse)
  })
_sym_db.RegisterMessage(ListClientConnectionsResponse)

CloseConnectionRequest = _reflection.GeneratedProtocolMessageType('CloseConnectionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLOSECONNECTIONREQUEST,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.CloseConnectionRequest)
  })
_sym_db.RegisterMessage(CloseConnectionRequest)

RouteInfo = _reflection.GeneratedProtocolMessageType('RouteInfo', (_message.Message,), {
  'DESCRIPTOR' : _ROUTEINFO,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.RouteInfo)
  })
_sym_db.RegisterMessage(RouteInfo)

GetGeneralInfoResponse = _reflection.GeneratedProtocolMessageType('GetGeneralInfoResponse', (_message.Message,), {

  'PluginInfo' : _reflection.GeneratedProtocolMessageType('PluginInfo', (_message.Message,), {
    'DESCRIPTOR' : _GETGENERALINFORESPONSE_PLUGININFO,
    '__module__' : 'yamcs.protobuf.web.general_service_pb2'
    # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetGeneralInfoResponse.PluginInfo)
    })
  ,
  'DESCRIPTOR' : _GETGENERALINFORESPONSE,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetGeneralInfoResponse)
  })
_sym_db.RegisterMessage(GetGeneralInfoResponse)
_sym_db.RegisterMessage(GetGeneralInfoResponse.PluginInfo)

ClientConnectionInfo = _reflection.GeneratedProtocolMessageType('ClientConnectionInfo', (_message.Message,), {

  'HttpRequestInfo' : _reflection.GeneratedProtocolMessageType('HttpRequestInfo', (_message.Message,), {
    'DESCRIPTOR' : _CLIENTCONNECTIONINFO_HTTPREQUESTINFO,
    '__module__' : 'yamcs.protobuf.web.general_service_pb2'
    # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo)
    })
  ,
  'DESCRIPTOR' : _CLIENTCONNECTIONINFO,
  '__module__' : 'yamcs.protobuf.web.general_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ClientConnectionInfo)
  })
_sym_db.RegisterMessage(ClientConnectionInfo)
_sym_db.RegisterMessage(ClientConnectionInfo.HttpRequestInfo)


DESCRIPTOR._options = None

_GENERALAPI = _descriptor.ServiceDescriptor(
  name='GeneralApi',
  full_name='yamcs.protobuf.web.GeneralApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=1178,
  serialized_end=1625,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetGeneralInfo',
    full_name='yamcs.protobuf.web.GeneralApi.GetGeneralInfo',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_GETGENERALINFORESPONSE,
    serialized_options=b'\212\222\003\006\n\004/api',
  ),
  _descriptor.MethodDescriptor(
    name='ListRoutes',
    full_name='yamcs.protobuf.web.GeneralApi.ListRoutes',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LISTROUTESRESPONSE,
    serialized_options=b'\212\222\003\r\n\013/api/routes',
  ),
  _descriptor.MethodDescriptor(
    name='ListClientConnections',
    full_name='yamcs.protobuf.web.GeneralApi.ListClientConnections',
    index=2,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LISTCLIENTCONNECTIONSRESPONSE,
    serialized_options=b'\212\222\003\022\n\020/api/connections',
  ),
  _descriptor.MethodDescriptor(
    name='CloseConnection',
    full_name='yamcs.protobuf.web.GeneralApi.CloseConnection',
    index=3,
    containing_service=None,
    input_type=_CLOSECONNECTIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\212\222\003\027\"\025/api/connections/{id}',
  ),
])
_sym_db.RegisterServiceDescriptor(_GENERALAPI)

DESCRIPTOR.services_by_name['GeneralApi'] = _GENERALAPI

# @@protoc_insertion_point(module_scope)



import functools
import socket
import threading

from yamcs.core.exceptions import YamcsError
from yamcs.core.futures import WebSocketSubscriptionFuture
from yamcs.core.helpers import adapt_name_for_rest, to_isostring
from yamcs.core.subscriptions import WebSocketSubscriptionManager
from yamcs.protobuf import yamcs_pb2
from yamcs.protobuf.alarms import alarms_pb2
from yamcs.protobuf.archive import archive_pb2
from yamcs.protobuf.mdb import mdb_pb2
from yamcs.protobuf.processing import processing_pb2
from yamcs.protobuf.pvalue import pvalue_pb2
from yamcs.protobuf.web import websocket_pb2
from yamcs.tmtc.model import (AlarmUpdate, Calibrator, CommandHistory,
                              IssuedCommand, MonitoredCommand, ParameterData,
                              ParameterValue, _parse_alarm)


class SequenceGenerator(object):
    """Static atomic counter."""
    _counter = 0
    _lock = threading.Lock()

    @classmethod
    def next(cls):
        with cls._lock:
            cls._counter += 1
            return cls._counter


def _wrap_callback_parse_parameter_data(subscription, on_data, message):
    """
    Wraps an (optional) user callback to parse ParameterData
    from a WebSocket data message
    """
    if message.type == message.REPLY:
        data = websocket_pb2.ParameterSubscriptionResponse()
        data.ParseFromString(message.reply.data)
        subscription.subscription_id = data.subscriptionId
    elif (message.type == message.DATA and
          message.data.type == yamcs_pb2.PARAMETER):
        parameter_data = ParameterData(getattr(message.data, 'parameterData'))
        #pylint: disable=protected-access
        subscription._process(parameter_data)
        if on_data:
            on_data(parameter_data)


def _wrap_callback_parse_cmdhist_data(subscription, on_data, message):
    """
    Wraps an (optional) user callback to parse CommandHistoryEntry
    from a WebSocket data message
    """
    if (message.type == message.DATA and
            message.data.type == yamcs_pb2.CMD_HISTORY):
        entry = getattr(message.data, 'command')
        #pylint: disable=protected-access
        rec = subscription._process(entry)
        if on_data:
            on_data(rec)


def _wrap_callback_parse_alarm_data(subscription, on_data, message):
    """
    Wraps an (optional) user callback to parse Alarm data
    from a WebSocket data message
    """
    if (message.type == message.DATA and
            message.data.type == yamcs_pb2.ALARM_DATA):
        proto = getattr(message.data, 'alarmData')
        alarm_update = AlarmUpdate(proto)
        #pylint: disable=protected-access
        subscription._process(alarm_update)
        if on_data:
            on_data(alarm_update)


def _build_named_object_id(parameter):
    """
    Builds a NamedObjectId. This is a bit more complex than it really
    should be. In Python (for convenience) we allow the user to simply address
    entries by their alias via the NAMESPACE/NAME convention. Yamcs is not
    aware of this convention so we decompose it into distinct namespace and
    name fields.
    """
    named_object_id = yamcs_pb2.NamedObjectId()
    if parameter.startswith('/'):
        named_object_id.name = parameter
    else:
        parts = parameter.split('/', 1)
        if len(parts) < 2:
            raise ValueError('Failed to process {}. Use fully-qualified '
                             'XTCE names or, alternatively, an alias in '
                             'in the format NAMESPACE/NAME'.format(parameter))
        named_object_id.namespace = parts[0]
        named_object_id.name = parts[1]
    return named_object_id


def _build_named_object_ids(parameters):
    """Builds a list of NamedObjectId."""
    if isinstance(parameters, str):
        return [_build_named_object_id(parameters)]
    return [_build_named_object_id(parameter) for parameter in parameters]


def _build_value_proto(value):
    proto = yamcs_pb2.Value()
    if isinstance(value, bool):
        proto.type = proto.BOOLEAN
        proto.booleanValue = value
    elif isinstance(value, float):
        proto.type = proto.FLOAT
        proto.floatValue = value
    elif isinstance(value, int) and value > 2147483647:
        proto.type = proto.SINT64
        proto.sint64Value = value
    elif isinstance(value, int):
        proto.type = proto.SINT32
        proto.sint32Value = value
    elif isinstance(value, str):
        proto.type = proto.STRING
        proto.stringValue = value
    else:
        raise YamcsError('Unrecognized type')
    return proto


def _set_range(ar, range, level):
    ar.level = level
    if range[0]:
        ar.minExclusive = range[0]
    if range[1]:
        ar.maxExclusive = range[1]


def _add_alarms(alarm_info, watch, warning, distress, critical, severe, min_violations):
    alarm_info.minViolations = min_violations

    if watch:
        ar = alarm_info.staticAlarmRange.add()
        _set_range(ar, watch, mdb_pb2.WATCH)
    if warning:
        ar = alarm_info.staticAlarmRange.add()
        _set_range(ar, warning, mdb_pb2.WARNING)
    if distress:
        ar = alarm_info.staticAlarmRange.add()
        _set_range(ar, distress, mdb_pb2.DISTRESS)
    if critical:
        ar = alarm_info.staticAlarmRange.add()
        _set_range(ar, critical, mdb_pb2.CRITICAL)
    if severe:
        ar = alarm_info.staticAlarmRange.add()
        _set_range(ar, severe, mdb_pb2.SEVERE)


def _add_calib(calib_info, type, data):
    type = type.lower()
    if type == Calibrator.POLYNOMIAL:
        calib_info.type = mdb_pb2.CalibratorInfo.POLYNOMIAL
        calib_info.polynomialCalibrator.coefficient.extend(data)
    elif type == Calibrator.SPLINE:
        calib_info.type = mdb_pb2.CalibratorInfo.SPLINE
        spline = mdb_pb2.SplineCalibratorInfo()
        for p in data:
            spi = spline.point.add()
            spi.raw = p[0]
            spi.calibrated = p[1]
    else:
        raise YamcsError('Unrecognized type')


class CommandHistorySubscription(WebSocketSubscriptionFuture):
    """
    Local object providing access to command history updates.

    This object buffers all received command history. This is needed
    to stitch together incremental command history events.

    If you expect to receive a lot of command history updates
    you should periodically clear local cache via ``clear_cache()``.
    In future work, we may add automated buffer management within
    configurable watermarks.

    .. warning::
        If command history updates are received for commands
        that are not currently in the local cache, the returned
        information may be incomplete.
    """

    @staticmethod
    def _cache_key(cmd_id):
        """commandId is a tuple. Convert to the equivalent string identifier."""
        return '{}-{}-{}'.format(
            cmd_id.generationTime, cmd_id.origin, cmd_id.sequenceNumber)

    def __init__(self, manager):
        super(CommandHistorySubscription, self).__init__(manager)
        self._cache = {}

    def clear_cache(self):
        """
        Clears local command history cache.
        """
        self._cache = {}

    def get_command_history(self, issued_command):
        """
        Gets locally cached CommandHistory for the specified command.

        :param .IssuedCommand issued_command: object representing a
                                              previously issued command.
        :rtype: .CommandHistory
        """
        #pylint: disable=protected-access
        if issued_command.id in self._cache:
            return self._cache[issued_command.id]
        return None

    def _process(self, entry):
        key = self._cache_key(entry.commandId)
        if key in self._cache:
            cmdhist = self._cache[key]
            #pylint: disable=protected-access
            cmdhist._update(entry.attr)
        else:
            cmdhist = CommandHistory(entry)
            self._cache[key] = cmdhist

        return cmdhist


class ParameterSubscription(WebSocketSubscriptionFuture):
    """
    Local object representing a subscription of zero or more parameters.

    A subscription object stores the last received value of each
    subscribed parameter.
    """

    def __init__(self, manager):
        super(ParameterSubscription, self).__init__(manager)

        self.value_cache = {}
        """Value cache keyed by parameter name."""

        self.delivery_count = 0
        """The number of parameter deliveries."""

        # The actual subscription_id is set async after server reply
        self.subscription_id = -1
        """Subscription number assigned by the server. This is set async,
        so may not be immediately available."""

    def add(self,
            parameters,
            abort_on_invalid=True,
            send_from_cache=True):
        """
        Add one or more parameters to this subscription.

        :param parameters: Parameter(s) to be added
        :type parameters: Union[str, str[]]
        :param bool abort_on_invalid: If ``True`` one invalid parameter
                                      means any other parameter in the
                                      request will also not be added
                                      to the subscription.
        :param bool send_from_cache: If ``True`` the last processed parameter
                                     value is sent from parameter cache.
                                     When ``False`` only newly processed
                                     parameters are received.
        """

        # Verify that we already know our assigned subscription_id
        assert self.subscription_id != -1

        if not parameters:
            return

        options = websocket_pb2.ParameterSubscriptionRequest()
        options.subscriptionId = self.subscription_id
        options.abortOnInvalid = abort_on_invalid
        options.sendFromCache = send_from_cache
        options.id.extend(_build_named_object_ids(parameters))

        self._manager.send('subscribe', options)

    def remove(self, parameters):
        """
        Remove one or more parameters from this subscription.

        :param parameters: Parameter(s) to be removed
        :type parameters: Union[str, str[]]
        """

        # Verify that we already know our assigned subscription_id
        assert self.subscription_id != -1

        if not parameters:
            return

        options = websocket_pb2.ParameterSubscriptionRequest()
        options.subscriptionId = self.subscription_id
        options.id.extend(_build_named_object_ids(parameters))

        self._manager.send('unsubscribe', options)

    def get_value(self, parameter):
        """
        Returns the last value of a specific parameter from local cache.

        :rtype: .ParameterValue
        """
        return self.value_cache[parameter]

    def _process(self, parameter_data):
        self.delivery_count += 1
        for pval in parameter_data.parameters:
            self.value_cache[pval.name] = pval


class CommandConnection(WebSocketSubscriptionFuture):
    """
    Local object providing access to the acknowledgment progress
    of command updates.

    Only commands issued from this object are monitored.
    """

    def __init__(self, manager, client):
        super(CommandConnection, self).__init__(manager)
        self._cmdhist_cache = {}
        self._command_cache = {}
        self._client = client

    def issue(self, command, args=None, dry_run=False, comment=None, verification=None):
        """
        Issue the given command

        :param str command: Either a fully-qualified XTCE name or an alias in the
                            format ``NAMESPACE/NAME``.
        :param dict args: Named arguments (if the command requires these)
        :param bool dry_run: If ``True`` the command is not actually issued. This
                             can be used to test if the server would generate
                             errors when preparing the command (for example
                             because an argument is missing).
        :param str comment: Comment attached to the command.
        :param .VerificationConfig verification: Overrides to the default
                                                 verification handling of this
                                                 command.
        :return: An object providing access to properties of the newly issued
                 command and updated according to command history updates.
        :rtype: .MonitoredCommand
        """
        issued_command = self._client.issue_command(command, args, dry_run, comment, verification)
        #pylint: disable=protected-access
        command = MonitoredCommand(issued_command._proto, self._client)

        self._command_cache[command.id] = command

        # It may be that we already received some cmdhist updates
        # before the http response returned.
        if command.id in self._cmdhist_cache:
            cmdhist = self._cmdhist_cache[command.id]
            command._process_cmdhist(cmdhist)

        return command

    #pylint: disable=protected-access
    def _process(self, entry):
        # TODO would be nice if the server gave this.
        command_id = '{}-{}-{}'.format(
            entry.commandId.generationTime, entry.commandId.origin,
            entry.commandId.sequenceNumber)

        if command_id in self._cmdhist_cache:
            cmdhist = self._cmdhist_cache[command_id]
            cmdhist._update(entry.attr)
        else:
            cmdhist = CommandHistory(entry)
            self._cmdhist_cache[command_id] = cmdhist

        command = self._command_cache.get(command_id)
        if command:
            command._process_cmdhist(cmdhist)

        return cmdhist


class AlarmSubscription(WebSocketSubscriptionFuture):
    """
    Local object representing an alarm subscription.

    A subscription object stores the currently active
    alarms.
    """

    def __init__(self, manager):
        super(AlarmSubscription, self).__init__(manager)

        self._cache = {}
        """Value cache keyed by alarm name."""

    def get_alarm(self, name):
        """
        Returns the alarm state associated with a specific named
        alarm from local cache.

        :param str name: Fully-qualified name
        :rtype: .Alarm
        """
        return self._cache[name]

    def list_alarms(self):
        """
        Returns a snapshot of all active alarms.

        :rtype: .Alarm[]
        """
        return [self._cache[k] for k in self._cache]

    def _process(self, alarm_update):
        alarm = alarm_update.alarm
        if alarm.processOK and not alarm.triggered and alarm.acknowledged:
            del self._cache[alarm.name]
        else:
            self._cache[alarm.name] = alarm


class ProcessorClient(object):
    """Client object that groups operations linked to a specific processor."""

    def __init__(self, client, instance, processor):
        super(ProcessorClient, self).__init__()
        self._client = client
        self._instance = instance
        self._processor = processor

    def get_parameter_value(self, parameter, from_cache=True, timeout=10):
        """
        Retrieve the current value of the specified parameter.

        :param str parameter: Either a fully-qualified XTCE name or an alias in the
                              format ``NAMESPACE/NAME``.
        :param bool from_cache: If ``False`` this call will block until a
                                fresh value is received on the processor.
                                If ``True`` the server returns the latest
                                value instead (which may be ``None``).
        :param float timeout: The amount of seconds to wait for a fresh value.
                              (ignored if ``from_cache=True``).
        :rtype: .ParameterValue
        """
        params = {
            'fromCache': from_cache,
            'timeout': int(timeout * 1000),
        }
        parameter = adapt_name_for_rest(parameter)
        url = '/processors/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        response = self._client.get_proto(url, params=params)
        proto = pvalue_pb2.ParameterValue()
        proto.ParseFromString(response.content)

        # Server returns ParameterValue with only 'id' set if no
        # value existed. Convert this to ``None``.
        if proto.HasField('rawValue') or proto.HasField('engValue'):
            return ParameterValue(proto)
        return None

    def get_parameter_values(self, parameters, from_cache=True, timeout=10):
        """
        Retrieve the current value of the specified parameter.

        :param str[] parameters: List of parameter names. These may be
                                 fully-qualified XTCE name or an alias
                                 in the format ``NAMESPACE/NAME``.
        :param bool from_cache: If ``False`` this call will block until
                                fresh values are received on the processor.
                                If ``True`` the server returns the latest
                                value instead (which may be ``None``).
        :param float timeout: The amount of seconds to wait for a fresh
                              values (ignored if ``from_cache=True``).
        :return: A list that matches the length and order of the requested
                 list of parameters. Each entry contains either the
                 returned parameter value, or ``None``.
        :rtype: .ParameterValue[]
        """
        req = processing_pb2.BatchGetParameterValuesRequest()
        req.id.extend(_build_named_object_ids(parameters))
        req.fromCache = from_cache
        req.timeout = int(timeout * 1000)
        url = '/processors/{}/{}/parameters:batchGet'.format(
            self._instance, self._processor)
        response = self._client.post_proto(url, data=req.SerializeToString())
        proto = processing_pb2.BatchGetParameterValuesResponse()
        proto.ParseFromString(response.content)

        pvals = []
        for parameter_id in req.id:
            match = None
            for pval in proto.value:
                if pval.id == parameter_id:
                    match = pval
                    break
            pvals.append(ParameterValue(match) if match else None)
        return pvals

    def set_parameter_value(self, parameter, value):
        """
        Sets the value of the specified parameter.

        :param str parameter: Either a fully-qualified XTCE name or an alias in the
                              format ``NAMESPACE/NAME``.
        :param value: The value to set
        """
        parameter = adapt_name_for_rest(parameter)
        url = '/processors/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        req = _build_value_proto(value)
        self._client.put_proto(url, data=req.SerializeToString())

    def set_parameter_values(self, values):
        """
        Sets the value of multiple  parameters.

        :param dict values: Values keyed by parameter name. This name can be either
                            a fully-qualified XTCE name or an alias in the format
                            ``NAMESPACE/NAME``.
        """
        req = processing_pb2.BatchSetParameterValuesRequest()
        for key in values:
            item = req.request.add()
            item.id.MergeFrom(_build_named_object_id(key))
            item.value.MergeFrom(_build_value_proto(values[key]))
        url = '/processors/{}/{}/parameters:batchSet'.format(
            self._instance, self._processor)
        self._client.post_proto(url, data=req.SerializeToString())

    def issue_command(self, command, args=None, dry_run=False, comment=None,
                      verification=None, attributes=None):
        """
        Issue the given command

        :param str command: Either a fully-qualified XTCE name or an alias in the
                            format ``NAMESPACE/NAME``.
        :param dict args: named arguments (if the command requires these)
        :param bool dry_run: If ``True`` the command is not actually issued. This
                             can be used to test if the server would generate
                             errors when preparing the command (for example
                             because an argument is missing).
        :param str comment: Comment attached to the command.
        :param .VerificationConfig verification: Overrides to the default
                                                 verification handling of this
                                                 command.
        :param dict attributes: named extra attributes that will be added to the command history.
                                Might be used by the processor or data links to change
                                the way the command is processed.
                                For example when commanding over a frame link where the COP1 protocol is used,
                                the cop1Bypass attribute can be used to by-pass the COP1 stack (the command will be sent into a so called Type-BD frame)
        :return: An object providing access to properties of the newly issued
                 command.
        :rtype: .IssuedCommand
        """
        req = processing_pb2.IssueCommandRequest()
        req.sequenceNumber = SequenceGenerator.next()
        req.dryRun = dry_run
        if comment:
            req.comment = comment
        if args:
            for key in args:
                assignment = req.assignment.add()
                assignment.name = key
                assignment.value = str(args[key])

        if verification:
            if verification._disable_all:
                req.commandOptions.disableVerifiers = True
            else:
                for verifier in verification._disabled:
                    req.commandOptions.verifierConfig[verifier].disable = True
                for verifier in verification._check_windows:
                    window = verification._check_windows[verifier]
                    if window['start']:
                        start = int(window['start'] * 1000)
                        req.commandOptions.verifierConfig[verifier].checkWindow.timeToStartChecking = start
                    if window['stop']:
                        stop = int(window['stop'] * 1000)
                        req.commandOptions.verifierConfig[verifier].checkWindow.timeToStopChecking = stop

        if attributes:
            for key in attributes:
                attr = req.attribute.add()
                attr.name = key
                attr.value.MergeFrom(_build_value_proto(attributes[key]))

        command = adapt_name_for_rest(command)
        url = '/processors/{}/{}/commands{}'.format(
            self._instance, self._processor, command)
        response = self._client.post_proto(url, data=req.SerializeToString())
        proto = processing_pb2.IssueCommandResponse()
        proto.ParseFromString(response.content)
        return IssuedCommand(proto, self)

    def list_alarms(self, start=None, stop=None):
        """
        Lists the active alarms.

        Remark that this does not query the archive. Only active alarms on the
        current processor are returned.

        :param ~datetime.datetime start: Minimum trigger time of the returned alarms (inclusive)
        :param ~datetime.datetime stop: Maximum trigger time of the returned alarms (exclusive)
        :rtype: ~collections.Iterable[.Alarm]
        """
        # TODO implement continuation token on server
        params = {
            'order': 'asc'
        }
        if start is not None:
            params['start'] = to_isostring(start)
        if stop is not None:
            params['stop'] = to_isostring(stop)
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        url = '/processors/{}/{}/alarms'.format(self._instance, self._processor)
        response = self._client.get_proto(path=url, params=params)
        message = alarms_pb2.ListAlarmsResponse()
        message.ParseFromString(response.content)
        alarms = getattr(message, 'alarms')
        return iter([_parse_alarm(alarm) for alarm in alarms])

    def set_default_calibrator(self, parameter, type, data):  # pylint: disable=W0622
        """
        Apply a calibrator while processing raw values of the specified
        parameter. If there is already a default calibrator associated
        to this parameter, that calibrator gets replaced.

        .. note::

            Contextual calibrators take precedence over the default calibrator
            See :meth:`set_calibrators` for setting contextual calibrators.

        Two types of calibrators can be applied:

        * Polynomial calibrators apply a polynomial expression of the form:
          `y = a + bx + cx^2 + ...`.

          The `data` argument must be an array of floats ``[a, b, c, ...]``.

        * Spline calibrators interpolate the raw value between a set of points
          which represent a linear curve.

          The `data` argument must be an array of ``[x, y]`` points.

        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        :param str type: One of ``polynomial`` or ``spline``.
        :param data: Calibration definition for the selected type.
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.SET_DEFAULT_CALIBRATOR
        if type:
            _add_calib(req.defaultCalibrator, type, data)

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def set_calibrators(self, parameter, calibrators):
        """
        Apply an ordered set of calibrators for the specified parameter.
        This replaces existing calibrators (if any).

        Each calibrator may have a context, which indicates when it its
        effects may be applied. Only the first matching calibrator is
        applied.

        A calibrator with context ``None`` is the *default* calibrator.
        There can be only one such calibrator, and is always applied at
        the end when no other contextual calibrator was applicable.

        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        :param .Calibrator[] calibrators: List of calibrators (either contextual or not)
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.SET_CALIBRATORS
        for c in calibrators:
            if c.context:
                context_calib = req.contextCalibrator.add()
                context_calib.context = c.context
                calib_info = context_calib.calibrator
            else:
                calib_info = req.defaultCalibrator

            _add_calib(calib_info, c.type, c.data)

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def clear_calibrators(self, parameter):
        """
        Removes all calibrators for the specified parameter.
        """
        self.set_default_calibrator(parameter, None, None)
        self.set_calibrators(parameter, [])

    def reset_calibrators(self, parameter):
        """
        Reset all calibrators for the specified parameter to their original MDB value.
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.RESET_CALIBRATORS

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def set_default_alarm_ranges(self, parameter, watch=None, warning=None,
                                 distress=None, critical=None, severe=None,
                                 min_violations=1):
        """
        Generate out-of-limit alarms for a parameter using the specified
        alarm ranges.

        This replaces any previous default alarms on this parameter.

        .. note::

            Contextual range sets take precedence over the default alarm
            ranges. See :meth:`set_alarm_range_sets` for setting contextual
            range sets.

        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        :param (float,float) watch: Range expressed as a tuple ``(lo, hi)``
                                where lo and hi are assumed exclusive.
        :param (float,float) warning: Range expressed as a tuple ``(lo, hi)``
                                  where lo and hi are assumed exclusive.
        :param (float,float) distress: Range expressed as a tuple ``(lo, hi)``
                                   where lo and hi are assumed exclusive.
        :param (float,float) critical: Range expressed as a tuple ``(lo, hi)``
                                   where lo and hi are assumed exclusive.
        :param (float,float) severe: Range expressed as a tuple ``(lo, hi)``
                                 where lo and hi are assumed exclusive.
        :param int min_violations: Minimum violations before an alarm is
                                   generated.
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.SET_DEFAULT_ALARMS
        if watch or warning or distress or critical or severe:
            _add_alarms(req.defaultAlarm, watch, warning, distress, critical,
                        severe, min_violations)

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def set_alarm_range_sets(self, parameter, sets):
        """
        Apply an ordered list of alarm range sets for the specified parameter.
        This replaces existing alarm sets (if any).

        Each RangeSet may have a context, which indicates when
        its effects may be applied. Only the first matching set is
        applied.

        A RangeSet with context ``None`` represents the *default* set of
        alarm ranges.  There can be only one such set, and it is always
        applied at the end when no other set of contextual ranges is
        applicable.

        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        :param .RangeSet[] sets: List of range sets (either contextual or not)
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.SET_ALARMS
        for rs in sets:
            if rs.context:
                context_alarm = req.contextAlarm.add()
                context_alarm.context = rs.context
                alarm_info = context_alarm.alarm
            else:
                alarm_info = req.defaultAlarm

            _add_alarms(alarm_info, rs.watch, rs.warning, rs.distress, rs.critical,
                        rs.severe, rs.min_violations)

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def clear_alarm_ranges(self, parameter):
        """
        Removes all alarm limits for the specified parameter.
        """
        self.set_default_alarm_ranges(parameter)
        self.set_alarm_range_sets(parameter, [])

    def reset_alarm_ranges(self, parameter):
        """
        Reset all alarm limits for the specified parameter to their original MDB value.
        """
        req = mdb_pb2.UpdateParameterRequest()
        req.action = mdb_pb2.UpdateParameterRequest.RESET_ALARMS

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/parameters{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def acknowledge_alarm(self, alarm, comment=None):
        """
        Acknowledges a specific alarm.

        :param alarm: Alarm instance
        :type alarm: :class:`.Alarm`
        :param str comment: Optional comment to associate with the state
                            change.
        """
        name = adapt_name_for_rest(alarm.name)
        url = '/processors/{}/{}/alarms{}/{}'.format(
            self._instance, self._processor, name, alarm.sequence_number)
        req = alarms_pb2.EditAlarmRequest()
        req.state = 'acknowledged'
        if comment is not None:
            req.comment = comment
        self._client.patch_proto(url, data=req.SerializeToString())

    def unshelve_alarm(self, alarm, comment=None):
        """
        Unshelve an alarm.

        :param alarm: Alarm instance
        :type alarm: :class:`.Alarm`
        :param str comment: Optional comment to associate with the state
                            change.
        """
        name = adapt_name_for_rest(alarm.name)
        url = '/processors/{}/{}/alarms{}/{}'.format(
            self._instance, self._processor, name, alarm.sequence_number)
        req = alarms_pb2.EditAlarmRequest()
        req.state = 'unshelved'
        self._client.patch_proto(url, data=req.SerializeToString())

    def shelve_alarm(self, alarm, comment=None):
        """
        Shelve an alarm.

        :param alarm: Alarm instance
        :type alarm: :class:`.Alarm`
        :param str comment: Optional comment to associate with the state
                            change.
        """
        name = adapt_name_for_rest(alarm.name)
        url = '/processors/{}/{}/alarms{}/{}'.format(
            self._instance, self._processor, name, alarm.sequence_number)
        req = alarms_pb2.EditAlarmRequest()
        req.state = 'shelved'
        if comment is not None:
            req.comment = comment
        self._client.patch_proto(url, data=req.SerializeToString())

    def clear_alarm(self, alarm, comment=None):
        """
        Clear an alarm.

        .. note::
            If the reason that caused the alarm is still present, a new
            alarm instance will be generated.

        :param alarm: Alarm instance
        :type alarm: :class:`.Alarm`
        :param str comment: Optional comment to associate with the state
                            change.
        """
        name = adapt_name_for_rest(alarm.name)
        url = '/processors/{}/{}/alarms{}/{}'.format(
            self._instance, self._processor, name, alarm.sequence_number)
        req = alarms_pb2.EditAlarmRequest()
        req.state = 'cleared'
        if comment is not None:
            req.comment = comment
        self._client.patch_proto(url, data=req.SerializeToString())

    def create_command_connection(self, on_data=None, timeout=60):
        """
        Creates a connection for issuing multiple commands and
        following up on their acknowledgment progress.

        .. note::
            This is a convenience method that merges the functionalities
            of :meth:`create_command_history_subscription` with those of
            :meth:`issue_command`.

        :param on_data: Function that gets called with  :class:`.CommandHistory`
                        updates. Only commands issued from this connection are
                        reported.
        :param timeout: The amount of seconds to wait for the request to
                        complete.
        :type timeout: float
        :return: Future that can be used to manage the background websocket
                 subscription
        :rtype: .CommandConnection
        """
        options = websocket_pb2.CommandHistorySubscriptionRequest()
        options.ignorePastCommands = True

        manager = WebSocketSubscriptionManager(
            self._client, resource='cmdhistory', options=options)

        # Represent subscription as a future
        subscription = CommandConnection(manager, self)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_cmdhist_data, subscription, on_data)

        manager.open(wrapped_callback, instance=self._instance,
                     processor=self._processor)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def create_command_history_subscription(self,
                                            on_data=None,
                                            timeout=60):
        """
        Create a new command history subscription.

        :param on_data: (Optional) Function that gets called with
                        :class:`.CommandHistory` updates.
        :param timeout: The amount of seconds to wait for the request to
                        complete.
        :type timeout: float
        :return: Future that can be used to manage the background websocket
                 subscription
        :rtype: .CommandHistorySubscription
        """
        options = websocket_pb2.CommandHistorySubscriptionRequest()
        options.ignorePastCommands = True

        manager = WebSocketSubscriptionManager(
            self._client, resource='cmdhistory', options=options)

        # Represent subscription as a future
        subscription = CommandHistorySubscription(manager)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_cmdhist_data, subscription, on_data)

        manager.open(wrapped_callback, instance=self._instance,
                     processor=self._processor)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def create_parameter_subscription(self,
                                      parameters,
                                      on_data=None,
                                      abort_on_invalid=True,
                                      update_on_expiration=False,
                                      send_from_cache=True,
                                      timeout=60):
        """
        Create a new parameter subscription.

        :param str[] parameters: Parameter names (or aliases).
        :param on_data: (Optional) Function that gets called with
                        :class:`.ParameterData` updates.
        :param bool abort_on_invalid: If ``True`` an error is generated when
                                      invalid parameters are specified.
        :param bool update_on_expiration: If ``True`` an update is received
                                          when a parameter value has become
                                          expired. This update holds the
                                          same value as the last known valid
                                          value, but with status set to
                                          ``EXPIRED``.
        :param bool send_from_cache: If ``True`` the last processed parameter
                                     value is sent from parameter cache.
                                     When ``False`` only newly processed
                                     parameters are received.
        :param float timeout: The amount of seconds to wait for the request
                              to complete.

        :return: A Future that can be used to manage the background websocket
                 subscription.
        :rtype: .ParameterSubscription
        """
        options = websocket_pb2.ParameterSubscriptionRequest()
        options.subscriptionId = -1  # This means 'create a new subscription'
        options.abortOnInvalid = abort_on_invalid
        options.updateOnExpiration = update_on_expiration
        options.sendFromCache = send_from_cache
        options.id.extend(_build_named_object_ids(parameters))

        manager = WebSocketSubscriptionManager(
            self._client, resource='parameter', options=options)

        # Represent subscription as a future
        subscription = ParameterSubscription(manager)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_parameter_data, subscription, on_data)

        manager.open(wrapped_callback, instance=self._instance,
                     processor=self._processor)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def create_alarm_subscription(self, on_data=None, timeout=60):
        """
        Create a new alarm subscription.

        :param on_data: (Optional) Function that gets called with
                        :class:`.AlarmUpdate` updates.
        :param float timeout: The amount of seconds to wait for the request
                              to complete.

        :return: A Future that can be used to manage the background websocket
                 subscription.
        :rtype: .AlarmSubscription
        """
        manager = WebSocketSubscriptionManager(self._client, resource='alarms')

        # Represent subscription as a future
        subscription = AlarmSubscription(manager)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_alarm_data, subscription, on_data)

        manager.open(wrapped_callback, instance=self._instance,
                     processor=self._processor)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def set_algorithm(self, parameter, text):
        """
        Change an algorithm text. Can only be peformed on JavaScript or Python algorithms.

        :param string text: new algorithm text (as it would appear in excel or XTCE)
        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        """
        req = mdb_pb2.UpdateAlgorithmRequest()
        req.action = mdb_pb2.UpdateAlgorithmRequest.SET
        req.algorithm.text = text

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/algorithms{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())

    def reset_algorithm(self, parameter):
        """
        Reset the algorithm text to its original definition from MDB

        :param str parameter: Either a fully-qualified XTCE name or an alias
                              in the format ``NAMESPACE/NAME``.
        """
        req = mdb_pb2.UpdateAlgorithmRequest()
        req.action = mdb_pb2.UpdateAlgorithmRequest.RESET

        parameter = adapt_name_for_rest(parameter)
        url = '/mdb/{}/{}/algorithms{}'.format(
            self._instance, self._processor, parameter)
        self._client.patch_proto(url, data=req.SerializeToString())
